#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <vector>

#include "pdlrp.h"
#include "pdlcount.h"
#include "sadacount.h"
#include "utils.h"

#include <rlcsa/docarray.h>
#include <rlcsa/misc/utils.h>

//--------------------------------------------------------------------------

void printUsage();

void bruteL(const RLCSA& rlcsa, const std::vector<pair_type>& ranges, usint& docc, double& seconds);
void bruteD(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size);
void pdlCount(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size);
void pdlList(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size);
void pdlRP(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size);
void sadaCount(const RLCSA& rlcsa, const std::string& base_name, char* parameters, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size);

//--------------------------------------------------------------------------

int
main(int argc, char** argv)
{
#ifdef VERBOSE
  std::cout << "Document counting test" << std::endl;
#endif
  if(argc < 3)
  {
    printUsage();
    return 1;
  }

  std::string base_name = argv[argc - 2];
  std::string patterns_name = argv[argc - 1];
#ifdef VERBOSE
  std::cout << "Base name: " << base_name << std::endl;
  std::cout << "Patterns: " << patterns_name << std::endl;
  std::cout << std::endl;
#endif

  // Index
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk())) { return 2; }
#ifdef VERBOSE
  rlcsa.printInfo();
  rlcsa.reportSize(true);
#endif

  // Patterns
  std::vector<std::string> rows;
  {
    std::ifstream pattern_file(patterns_name.c_str(), std::ios_base::binary);
    if(!pattern_file)
    {
      std::cerr << "Error opening pattern file!" << std::endl;
      return 3;
    }
    CSA::readRows(pattern_file, rows, true);
    pattern_file.close();
  }

  // find()
  std::vector<pair_type> ranges; ranges.reserve(rows.size());
  {
    usint total_size = 0, total_occ = 0;
    double start = CSA::readTimer();
    for(usint i = 0; i < rows.size(); i++)
    {
      pair_type result = rlcsa.count(rows[i]);
      total_size += rows[i].length(); total_occ += CSA::length(result);
      ranges.push_back(result);
    }
    double seconds = CSA::readTimer() - start;
    double megabytes = total_size / (double)CSA::MEGABYTE;
#ifdef VERBOSE
    std::cout << "Patterns:     " << rows.size() << " (" << megabytes << " MB)" << std::endl;
    std::cout << "Occurrences:  " << total_occ << std::endl;
    std::cout << "Find:         " << seconds << " seconds (" << (megabytes / seconds) << " MB/s, "
                                  << (rows.size() / seconds) << " patterns/s)" << std::endl;
    std::cout << std::endl;
#endif
  }

  // Algorithms
  for(int arg = 1; arg < argc - 2; arg++)
  {
    std::string algorithm;
    usint docc = 0, size = 0;
    double seconds = 0;
    switch(argv[arg][0])
    {
      case 'b':
        algorithm = "Brute-L";
        bruteL(rlcsa, ranges, docc, seconds);
        break;
      case 'B':
        algorithm = "Brute-D";
        bruteD(rlcsa, base_name, ranges, docc, seconds, size);
        break;
      case 'l':
        algorithm = "PDL-BC";
        pdlList(rlcsa, base_name, ranges, docc, seconds, size);
        break;
      case 'L':
        algorithm = "PDL-RP";
        pdlRP(rlcsa, base_name, ranges, docc, seconds, size);
        break;
      case 'p':
        algorithm = "PDL-count";
        pdlCount(rlcsa, base_name, ranges, docc, seconds, size);
        break;
      case 's':
        algorithm = "Sada-"; algorithm += argv[arg] + 1;
        sadaCount(rlcsa, base_name, argv[arg] + 1, ranges, docc, seconds, size);
        break;
      default:
        std::cerr << "count_docs: Invalid algorithm: " << argv[arg] << std::endl;
        break;
    }
    if(algorithm.length() > 0)
    {
#ifdef VERBOSE
      printResults(algorithm, seconds, docc, rows.size());
      printSize(algorithm, rlcsa.getSize(), rlcsa.reportSize(), size);
      std::cout << std::endl;
#else
      printCount(algorithm, rlcsa.getSize(), size, seconds, rows.size());
#endif
    }
  }

  return 0;
}

//--------------------------------------------------------------------------

void
printUsage()
{
  std::cout << "Usage: count_docs [algorithms] base_name patterns" << std::endl;
  std::cout << "  b   Brute force using locate()" << std::endl;
  std::cout << "  B   Brute force using D" << std::endl;
  std::cout << "  l   Precomputed document listing" << std::endl;
  std::cout << "  L   Precomputed document listing using RePair" << std::endl;
  std::cout << "  p   Precomputed document counting" << std::endl;
  std::cout << "  sV  Sadakane's document counting" << std::endl;
  std::cout << "  sVF Sadakane's document counting using a filter" << std::endl;
  std::cout << std::endl;
  std::cout << "V denotes an encoding for the bitvector and F for the filter. The alternatives are" << std::endl;
  std::cout << "P(lain), G(ap encoded), and R(un-length encoded)." << std::endl;
  std::cout << std::endl;
}

template<class M>
void
countWith(const M& method, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  size = method.reportSize();
  double start = CSA::readTimer();
  for(usint i = 0; i < ranges.size(); i++) { docc += method.count(ranges[i]); }
  seconds = CSA::readTimer() - start;
}

//--------------------------------------------------------------------------

void
bruteL(const RLCSA& rlcsa, const std::vector<pair_type>& ranges, usint& docc, double& seconds)
{
  double start = CSA::readTimer();
  for(usint i = 0; i < ranges.size(); i++) { docc += bruteForceDocCount(rlcsa, ranges[i]); }
  seconds = CSA::readTimer() - start;
}

void
bruteD(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  CSA::ReadBuffer* docarray = readDocArray(rlcsa, base_name);
  if(docarray == 0)
  {
    std::cerr << "count_docs: Cannot initialize DoclistSada!" << std::endl;
    return;
  }
  size = docarray->reportSize();
  double start = CSA::readTimer();
  for(usint i = 0; i < ranges.size(); i++) { docc += bruteForceDocCount(*docarray, ranges[i]); }
  seconds = CSA::readTimer() - start;
  delete docarray; docarray = 0;
}

void
pdlCount(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  PDLCount pdlcount(rlcsa, base_name);
  if(!(pdlcount.isOk()))
  {
    std::cerr << "count_docs: Cannot initialize PDL-count!" << std::endl;
    return;
  }
  countWith(pdlcount, ranges, docc, seconds, size);
}

void
pdlList(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  CSA::DocArray pdl(rlcsa, base_name, true);
  if(!(pdl.isOk()))
  {
    std::cerr << "count_docs: Cannot initialize PDL-BC!" << std::endl;
    return;
  }
  countWith(pdl, ranges, docc, seconds, size);
}

void
pdlRP(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  PDLRP rp(rlcsa, base_name, false);
  if(!(rp.isOk()))
  {
    std::cerr << "count_docs: Cannot initialize PDL-RP!" << std::endl;
    return;
  }
  countWith(rp, ranges, docc, seconds, size);
}

//--------------------------------------------------------------------------

const static usint ENC_NONE  = 0x0;
const static usint ENC_PLAIN = 0x1;
const static usint ENC_GAP   = 0x2;
const static usint ENC_RLE   = 0x3;

const static usint ENC_SHIFT = 4;
const static usint ENC_PLAIN_PLAIN = 0x11;
const static usint ENC_PLAIN_GAP   = 0x12;
const static usint ENC_PLAIN_RLE   = 0x13;
const static usint ENC_GAP_PLAIN   = 0x21;
const static usint ENC_GAP_GAP     = 0x22;
const static usint ENC_GAP_RLE     = 0x23;
const static usint ENC_RLE_PLAIN   = 0x31;
const static usint ENC_RLE_GAP     = 0x32;
const static usint ENC_RLE_RLE     = 0x33;

usint
identify(char encoding)
{
  switch(encoding)
  {
    case 'P': return ENC_PLAIN;
    case 'G': return ENC_GAP;
    case 'R': return ENC_RLE;
    default:  return ENC_NONE;
  }
}

template<class V>
void
countWithSada(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  SadaCount<V> sada(rlcsa, base_name);
  if(!(sada.isOk()))
  {
    std::cerr << "count_docs: Cannot initialize SadaCount!" << std::endl;
    return;
  }
  countWith(sada, ranges, docc, seconds, size);
}

template<class V, class F>
void
countWithSada(const RLCSA& rlcsa, const std::string& base_name, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  SadaFilter<V, F> sada(rlcsa, base_name);
  if(!(sada.isOk()))
  {
    std::cerr << "count_docs: Cannot initialize SadaFilter!" << std::endl;
    return;
  }
  countWith(sada, ranges, docc, seconds, size);
}

void
sadaCount(const RLCSA& rlcsa, const std::string& base_name, char* parameters, const std::vector<pair_type>& ranges, usint& docc, double& seconds, usint& size)
{
  usint len = strlen(parameters);
  if(len == 1)  // No filter.
  {
    usint encoding = identify(parameters[0]);
    switch(encoding)
    {
      case ENC_PLAIN:
        countWithSada<CSA::SuccinctVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_GAP:
        countWithSada<CSA::DeltaVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_RLE:
        countWithSada<CSA::RLEVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
    }
  }
  else if(len == 2) // Using a filter.
  {
    usint encoding = (identify(parameters[0]) << ENC_SHIFT) | identify(parameters[1]);
    switch(encoding)
    {
      case ENC_PLAIN_PLAIN:
        countWithSada<CSA::SuccinctVector, CSA::SuccinctVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_PLAIN_GAP:
        countWithSada<CSA::SuccinctVector, CSA::DeltaVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_PLAIN_RLE:
        countWithSada<CSA::SuccinctVector, CSA::RLEVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_GAP_PLAIN:
        countWithSada<CSA::DeltaVector, CSA::SuccinctVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_GAP_GAP:
        countWithSada<CSA::DeltaVector, CSA::DeltaVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_GAP_RLE:
        countWithSada<CSA::DeltaVector, CSA::RLEVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_RLE_PLAIN:
        countWithSada<CSA::RLEVector, CSA::SuccinctVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_RLE_GAP:
        countWithSada<CSA::RLEVector, CSA::DeltaVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
      case ENC_RLE_RLE:
        countWithSada<CSA::RLEVector, CSA::RLEVector>(rlcsa, base_name, ranges, docc, seconds, size);
        return;
    }
  }
  std::cerr << "count_docs: Invalid encoding for SadaCount: " << parameters << std::endl;
}

//--------------------------------------------------------------------------
