#include <algorithm>
#include <fstream>
#include <iostream>
#include <stack>

#include <rlcsa/bits/vectors.h>

#include "pdlcount.h"


//--------------------------------------------------------------------------

struct CountTreeNode;
void deleteNodes(CountTreeNode* root);

struct CountTreeNode
{
  union
  {
    usint               string_depth;
    std::vector<usint>* docs;
    usint               docc;
  };
  pair_type range;

  CountTreeNode* child;
  CountTreeNode* sibling;

  CountTreeNode()
  {
    this->string_depth = 0;
    this->range = pair_type(0, 0);
    this->child = this->sibling = 0;
  }

  // Deletion leaks memory if docs has been set.
  ~CountTreeNode()
  {
  }

  void addChild(CountTreeNode* node)
  {
    if(this->child == 0) { this->child = node; return; }
    CountTreeNode* temp = this->child;
    while(temp->sibling != 0) { temp = temp->sibling; }
    temp->sibling = node;
  }

  // Adds leaves to the sparse suffix tree when required.
  void addLeaves()
  {
    std::stack<CountTreeNode*> nodestack;
    nodestack.push(this);

    while(!(nodestack.empty()))
    {
      CountTreeNode* curr = nodestack.top(); nodestack.pop();
      if(curr->child == 0) { continue; }
      usint expect = curr->range.first;
      CountTreeNode* prev = 0;
      for(CountTreeNode* temp = curr->child; temp != 0; temp = temp->sibling)
      {
        while(expect < temp->range.first) { prev = curr->addLeaf(prev, temp, expect); expect++; }
        nodestack.push(temp);
        expect = temp->range.second + 1;
        prev = temp;
      }
      while(expect <= curr->range.second) { prev = curr->addLeaf(prev, 0, expect); expect++; }
    }
  }

  CountTreeNode* addLeaf(CountTreeNode* left, CountTreeNode* right, usint pos)
  {
    CountTreeNode* leaf = new CountTreeNode;
    leaf->range = pair_type(pos, pos);
    if(left == 0) { this->child = leaf; }
    else          { left->sibling = leaf; }
    leaf->sibling = right;
    return leaf;
  }

  void
  determineSize(usint& nodes, usint& leaves)
  {
    nodes = 0; leaves = 0;
    std::stack<CountTreeNode*> nodestack;
    nodestack.push(this);

    while(!(nodestack.empty()))
    {
      CountTreeNode* curr = nodestack.top(); nodestack.pop(); nodes++;
      if(curr->child == 0) { leaves++; }
      for(CountTreeNode* temp = curr->child; temp != 0; temp = temp->sibling) { nodestack.push(temp); }
    }
  }

  // Leaks memory if docs has been set.
  void computeStoredDocuments(usint* documents)
  {
    bool keep_children = false; // Stays false if all children have occ == docc.
    if(this->child == 0)
    {
      this->docs = new std::vector<usint>(documents + this->range.first, documents + this->range.second + 1);
      keep_children = true;
    }
    else
    {
      this->docs = new std::vector<usint>;
      for(CountTreeNode* curr = this->child; curr != 0; curr = curr->sibling)
      {
        this->docs->insert(this->docs->end(), curr->docs->begin(), curr->docs->end());
        curr->deleteDocs();
        if(curr->docc < CSA::length(curr->range)) { keep_children = true; }
      }
    }
    CSA::removeDuplicates(this->docs, (this->docs->size() >= PARALLEL_SORT_THRESHOLD));
    if(!keep_children) { deleteNodes(this->child); this->child = 0; }
  }

  // Replaces docs with docc.
  void deleteDocs()
  {
    usint temp = this->docs->size();
    delete this->docs; this->docc = temp;
  }

  const static usint PARALLEL_SORT_THRESHOLD = 4096;
};

// Deletes root, it's right siblings, and their subtrees.
// The reason for the existence of this function is that a recursive
// destructor may result in a stack overflow in very tall trees.
// This leaks memory if one of the nodes has docs set.
void
deleteNodes(CountTreeNode* root)
{
  std::stack<CountTreeNode*> nodestack;
  while(root != 0)
  {
    nodestack.push(root);
    CountTreeNode* temp = root->child; root->child = 0; root = temp;
  }
  while(!(nodestack.empty()))
  {
    root = nodestack.top();
    if(root->sibling == 0) { delete root; nodestack.pop(); }
    else
    {
      CountTreeNode* temp = root->sibling; root->sibling = 0; root = temp;
      while(root != 0)
      {
        nodestack.push(root);
        temp = root->child; root->child = 0; root = temp;
      }
    }
  }
}

//--------------------------------------------------------------------------

void
buildPDLCount(const std::string& base_name, bool print)
{
  // Load RLCSA.
  if(print) { std::cout << "Loading RLCSA..." << std::endl; }
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk()) || !(rlcsa.supportsLocate()))
  {
    std::cerr << "buildPDLCount(): Invalid RLCSA!" << std::endl;
    return;
  }
  rlcsa.printInfo();
  rlcsa.reportSize(true);

  // Timer start.
  double start = CSA::readTimer();

  // Build SA.
  if(print) { std::cout << "Building SA..." << std::endl; }
  usint size = rlcsa.getSize();
  usint* sa = new usint[size];
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < size; i += CSA::MEGABYTE)
  {
    rlcsa.locate(pair_type(i, std::min(i + CSA::MEGABYTE - 1, size - 1)), sa + i);
  }

  // Build LCP.
  if(print) { std::cout << "Building LCP..." << std::endl; }
  uint* lcp = new uint[size + 1]; lcp[size] = 0;
  CSA::PLCPVector* plcpvec = rlcsa.buildPLCP(16);
  CSA::PLCPVector::Iterator iter(*plcpvec);
  for(usint i = 0; i < size; i++)
  {
    lcp[i] = iter.select(sa[i]) - 2 * sa[i];
  }
  delete plcpvec; plcpvec = 0;

  // Convert SA to DA.
  if(print) { std::cout << "Building DA..." << std::endl; }
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < size; i += CSA::MEGABYTE)
  {
    rlcsa.getSequenceForPosition(sa + i, std::min(CSA::MEGABYTE, size - i));
  }

  // Build ST. Kind of.
  if(print) { std::cout << "Building ST..."; std::cout.flush(); }
  CountTreeNode* root = new CountTreeNode();
  root->string_depth = 0; root->range = pair_type(0, 0);
  std::stack<CountTreeNode*> nodestack; nodestack.push(root);
  CountTreeNode* prev = 0;
  for(usint i = 1; i <= size; i++)
  {
    usint left = i - 1;
    while(lcp[i] < nodestack.top()->string_depth)
    {
      nodestack.top()->range.second = i - 1;
      prev = nodestack.top(); nodestack.pop();
      left = prev->range.first;
      if(lcp[i] <= nodestack.top()->string_depth) { nodestack.top()->addChild(prev); prev = 0; }
    }
    if(lcp[i] > nodestack.top()->string_depth)
    {
      CountTreeNode* curr = new CountTreeNode();
      curr->string_depth = lcp[i]; curr->range = pair_type(left, left);
      if(prev != 0) { curr->addChild(prev); prev = 0; }
      nodestack.push(curr);
    }
  }
  while(!(nodestack.empty()))
  {
    prev = nodestack.top(); nodestack.pop();
    prev->range.second = size - 1;
  }
  delete[] lcp; lcp = 0;
  root->addLeaves();
  usint nodes = 0, leaves = 0;
  if(print)
  {
    root->determineSize(nodes, leaves);
    std::cout << " Size: " << nodes << " / " << size << std::endl;
  }

  // Build the sets of document ids.
  if(print) { std::cout << "Building sets..."; std::cout.flush(); }
  for(prev = root; prev != 0; prev = prev->child) { nodestack.push(prev); }
  while(!(nodestack.empty()))
  {
    CountTreeNode* curr = nodestack.top(); nodestack.pop();
    for(CountTreeNode* temp = curr->sibling; temp != 0; temp = temp->child) { nodestack.push(temp); }
    curr->computeStoredDocuments(sa);
  }
  delete[] sa; sa = 0;
  root->deleteDocs();
  root->determineSize(nodes, leaves);
  if(print) { std::cout << " " << nodes << " nodes, " << leaves << " leaves" << std::endl;}

  // Building the counting structure.
  if(print) { std::cout << "Building PDL-count..." << std::endl; }
  CSA::WriteBuffer ranges(size, 1); // Mark the beginning of each leaf node.
  CSA::Array::Encoder encoder(PDLCount::COUNT_BLOCK_SIZE);  // Document counts in preorder.
  sdsl::bit_vector tree(2 * nodes, 0);  // Balanced parenthesis tree representation.
  usint pos = 0;
  for(prev = root; prev != 0; prev = prev->child)
  {
    nodestack.push(prev);
    tree[pos] = 1; pos++;
    encoder.writeItem(prev->docc);
  }
  while(!(nodestack.empty()))
  {
    CountTreeNode* curr = nodestack.top(); nodestack.pop();
    if(curr->child == 0) { ranges.goToItem(curr->range.first); ranges.writeItem(1); }
    tree[pos] = 0; pos++;
    for(CountTreeNode* temp = curr->sibling; temp != 0; temp = temp->child)
    {
      nodestack.push(temp);
      tree[pos] = 1; pos++;
      encoder.writeItem(temp->docc);
    }
  }
  deleteNodes(root); root = 0;
  CSA::DeltaVector* leaf_ranges = CSA::encode<CSA::DeltaVector>(ranges, PDLCount::RANGE_BLOCK_SIZE, size);
  CSA::Array counts(encoder);

  // Timer stop.
  double total = CSA::readTimer() - start;

  // Writing results to a file.
  std::string output_name = base_name + PDLCount::EXTENSION;
  std::ofstream output(output_name.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "buildPDLCount(): Cannot open output file " << output_name << std::endl;
    return;
  }
  leaf_ranges->writeTo(output);
  counts.writeTo(output);
  tree.serialize(output);
  output.close();
  delete leaf_ranges; leaf_ranges = 0;

  if(print)
  {
    double megabytes = size / (double)CSA::MEGABYTE;
    std::cout << "Processed " << megabytes << " megabytes in " << total << " seconds (" << (megabytes / total) << " MB/s)." << std::endl;
    std::cout << "Memory: " << CSA::memoryUsage() << " kB" << std::endl;
    std::cout << std::endl;
  }
}

//--------------------------------------------------------------------------

const std::string PDLCount::EXTENSION = ".pdlcount";

PDLCount::PDLCount(const RLCSA& _rlcsa, const std::string& base_name) :
  rlcsa(_rlcsa), status(st_error),
  leaf_ranges(0), counts(0)
{
  if(!(this->rlcsa.isOk()))
  {
    std::cerr << "PDLCount::PDLCount(): Invalid RLCSA!" << std::endl;
    return;
  }

  std::string input_name = base_name + EXTENSION;
  std::ifstream input(input_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "PDLCount::PDLCount(): Cannot open input file " << input_name << std::endl;
    return;
  }

  this->leaf_ranges = new CSA::DeltaVector(input);
  this->counts = new CSA::Array(input);
  this->tree.load(input);
  input.close();

  sdsl::util::init_support(this->tree_rank, &(this->tree));
  sdsl::util::init_support(this->tree_select, &(this->tree));

  this->status = st_ok;
}

PDLCount::~PDLCount()
{
  delete this->leaf_ranges; this->leaf_ranges = 0;
  delete this->counts; this->counts = 0;
}

void
PDLCount::writeTo(const std::string& base_name) const
{
  if(!(this->isOk()))
  {
    std::cerr << "PDLCount::writeTo(): Status is not ok!" << std::endl;
    return;
  }

  std::string filename = base_name + EXTENSION;
  std::ofstream output(filename.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "PDLCount::writeTo(): Cannot open output file " << filename << std::endl;
    return;
  }
  this->leaf_ranges->writeTo(output);
  this->counts->writeTo(output);
  this->tree.serialize(output);
  output.close();
}

usint
PDLCount::reportSize() const
{
  usint bytes = sizeof(*this);
  if(this->leaf_ranges != 0) { bytes += this->leaf_ranges->reportSize(); }
  if(this->counts != 0) { bytes += this->counts->reportSize(); }
  bytes += sdsl::size_in_bytes(this->tree);
  bytes += sdsl::size_in_bytes(this->tree_rank);
  bytes += sdsl::size_in_bytes(this->tree_select);
  return bytes;
}

//--------------------------------------------------------------------------

usint
PDLCount::count(const std::string& pattern) const
{
  if(!(this->isOk())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  return this->count(sa_range);
}

usint
PDLCount::count(pair_type sa_range) const
{
  if(!(this->isOk()) || CSA::isEmpty(sa_range) || sa_range.second >= this->rlcsa.getSize()) { return 0; }

  CSA::DeltaVector::Iterator block_iter(*(this->leaf_ranges));
  pair_type first_block = block_iter.valueAfter(sa_range.first);
  if(first_block.first != sa_range.first) { return CSA::length(sa_range); }

  block_iter.valueBefore(sa_range.second);
  pair_type next_block = block_iter.nextValue();
  if(next_block.first != sa_range.second + 1) { return CSA::length(sa_range); }

  // The range of leaf nodes is [first_block.second, next_block.second - 1].
  // We convert the blocks into (start of run, rank at start).
  // Note that SDSL ranks are 0-based and select for pattern 01 returns the endpoint.
  CSA::Array::Iterator count_iter(*(this->counts));
  first_block.first = (first_block.second == 0 ? 0 : this->tree_select(first_block.second));
  first_block.second = this->tree_rank(first_block.first);
  // FIXME does this work with the last block in SDSL?
  next_block.first = (next_block.second == 0 ? 0 : this->tree_select(next_block.second));
  next_block.second = this->tree_rank(next_block.first);

  // Convert the blocks into (depth at the start of the run, rank at start).
  first_block.first = 2 * first_block.second + 1 - first_block.first;
  next_block.first = 2 * next_block.second + 1 - next_block.first;

  // Either the LCA is at the start of first_block, or its closing parenthesis is
  // just before next_block, depending of which of them is deeper.
  return count_iter.readItem(first_block.second +
    (first_block.first >= next_block.first ? 0 : next_block.first - first_block.first));
}

//--------------------------------------------------------------------------
